import ast
import inspect
from typing import Tuple, Type, Optional

from engage.parser import fetch_functions, func_summary

from astor import to_source


class Writer:

    def __init__(self):
        self.docs = []

    def update_arg(self, arg: Tuple[str, Optional[str]]):
        raise NotImplemented

    def update_kwargs(self, kwarg: Tuple[str, Optional[str]]):
        raise NotImplemented

    def update_vararg(self, vararg: [str, Optional[str]]):
        raise NotImplemented

    def update_default_arg(self, arg: Tuple[str, Optional[str], str]):
        raise NotImplemented

    def update_return(self, return_doc: Optional[str]):
        raise NotImplemented

    def digest(self, indent: int) -> str:
        split = "\n" + " " * indent
        return split + split.join(self.docs) + split


class Sphinx(Writer):

    def update_arg(self, arg: Tuple[str, Optional[str]]):
        if arg[0] == 'self':
            return
        self.docs.append(":param {}: ".format(arg[0]))
        if arg[1] is not None:
            self.docs.append(":type {}: {}".format(arg[0], arg[1].strip("\n() ")))
        else:
            self.docs.append(":type {}: ".format(arg[0]))

    def update_default_arg(self, arg: Tuple[str, Optional[str], str]):
        self.docs.append(":param {}:  (Default: {})".format(arg[0], arg[2].strip("\n() ")))
        if arg[1] is not None:
            self.docs.append(":type {}: {}".format(arg[0], arg[1].strip("\n() ")))
        else:
            self.docs.append(":type {}: ".format(arg[0]))

    def update_kwarg(self, kwarg: Tuple[str, Optional[str]]):
        self.docs.append(":param **{}: ".format(kwarg[0]))
        if kwarg[1] is not None:
            self.docs.append(":type **{}: {}".format(kwarg[0], kwarg[1].strip("\n() ")))
        else:
            self.docs.append(":type **{}: ".format(kwarg[0]))

    def update_return(self, return_doc: Optional[str]):
        self.docs.append(":return: ")
        if return_doc is not None:
            self.docs.append(":rtype: {}".format(return_doc))
        else:
            self.docs.append(":rtype: ")

    def update_vararg(self, vararg: [str, Optional[str]]):
        self.docs.append(":param *{}: ".format(vararg[0]))
        if vararg[1] is not None:
            self.docs.append(":type *{}: {}".format(vararg[0], vararg[1].strip("\n() ")))
        else:
            self.docs.append(":type *{}: ".format(vararg[0]))

    def update_docs(self, docs: str):
        cleaned = inspect.cleandoc(docs)
        self.docs = cleaned.split("\n") + [""] + self.docs


def update_module(contents: str, writer: Type[Writer] = Sphinx) -> str:
    func_data = map(func_summary, fetch_functions(contents))
    summary = []
    for function in func_data:
        author = writer()
        for name, value in function.items():
            if value and hasattr(author, "update_" + name):
                if not isinstance(value, list):
                    getattr(author, "update_" + name)(value)
                else:
                    for item in value:
                        getattr(author, "update_" + name)(item)
        summary.append(author.digest(function['indent']))

    func_generator = fetch_functions(contents, True)
    for doc, part in zip(summary, func_generator):
        if isinstance(part.body[0], ast.Expr) and isinstance(part.body[0].value, ast.Str):
            part.body[0].value.s = doc
        else:
            e = ast.Expr([ast.Str(doc)])
            part.body.insert(0, e)
    return to_source(next(func_generator))



from pathlib import Path

from engage.traverse import traverse, get_contents
from engage.writer import update_module, Sphinx, Writer


def engage_path(file_path: str, writer: Writer=Sphinx):
    for file in traverse(Path(file_path)):
        new_file = update_module(
            get_contents(str(file)),
            writer
        )
        with open(file, "w") as f:
            f.write(new_file)



import ast
from typing import Any, Dict, List, Tuple, Optional, Iterator, Union
import warnings

import astor


def fetch_functions(content: str, return_body: bool = False) -> Iterator[Union[ast.FunctionDef, ast.AsyncFunctionDef]]:
    body = ast.parse(content)
    for part in ast.walk(body):
        if isinstance(part, (ast.FunctionDef, ast.AsyncFunctionDef)):
            yield part
    if return_body:
        yield body


def has_return(func: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> bool:
    return any(isinstance(node, ast.Return) for node in ast.walk(func))


def get_return_doc(func: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> str:
    return astor.to_source(func.returns).strip() if func.returns else None


def get_args(func: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> List[Tuple[str, Optional[str]]]:
    return [(arg.arg, (astor.to_source(arg.annotation)) if arg.annotation else None) for arg in func.args.args]


def get_kwargs(func: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> List[Tuple[str, Optional[str], str]]:
    return [(arg.arg, (astor.to_source(arg.annotation) if arg.annotation else None), astor.to_source(default))
            for arg, default in zip(func.args.kwonlyargs, func.args.kw_defaults)]


def get_kwarg(func: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Tuple[Optional[str], Optional[str]]:
    return func.args.kwarg.arg, (astor.to_source(func.args.kwarg.annotation) if func.args.kwarg.annotation else None)


def get_vararg(func: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Tuple[Optional[str], Optional[str]]:
    return func.args.vararg.arg, (astor.to_source(func.args.vararg.annotation) if func.args.vararg.annotation else None)


def get_indent(func: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> int:
    sorted_parts = sorted(filter(lambda x: x.col_offset != -1, func.body), key=lambda x: x.col_offset)
    try:
        return sorted_parts[0].col_offset
    except IndexError:
        warnings.warn("Unable to identify indentation for function {}, make sure the function is more than just a "
                      "docstring. This can be fixed by added 'pass'".format(func.name), UserWarning)
        return 4


def get_doc(func: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Optional[str]:
    if func.body:
        if isinstance(func.body[0], ast.Expr) and isinstance(func.body[0].value, ast.Str):
            return func.body[0].value.s


def func_summary(func: Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Dict[str, Any]:
    return {
        "returns": has_return(func),
        "return_doc": get_return_doc(func),
        "vararg": get_vararg(func) if func.args.vararg else None,
        "kwarg": get_kwarg(func) if func.args.kwarg else None,
        "default_arg": get_kwargs(func),
        "arg": get_args(func),
        "docs": get_doc(func),
        "name": func.name,
        "indent": get_indent(func)
    }



from pathlib import Path
from typing import List
from os import listdir


def traverse(path: Path) -> List[Path]:
    if path.is_file():
        return [path]
    else:
        parts = listdir(str(path))
        if '__init__.py' in parts:
            return sum([traverse(path.joinpath(part)) for part in parts], [])
        return []


def get_contents(file_path: str):
    with open(file_path) as f:
        return f.read()



from argparse import ArgumentParser

from engage import engage_path


parser = ArgumentParser()
parser.add_argument("package", help="The file / directory to engage")


if __name__ == "__main__":
    args = parser.parse_args()
    engage_path(args.package)



from setuptools import setup


setup(
    name="engage",
    description="Sphinx autodoc templater",
    author="Nathan Zilora",
    author_email="zwork101@gmail.com",
    url="https://github.com/Zwork101/engage",
    license="MIT",
    packages=['engage'],
    version="1.0.1",
    install_requires=["astor"]
)
